extern "C" 
{
	#include <wdm.h> 
}

struct DEVICE_EXTENSION 
{
	UNICODE_STRING SymLinkName;	
	void* pBuffer;
	ULONG BufferSize;
} ;

void DriverUnload ( IN PDRIVER_OBJECT pDrvObj );
NTSTATUS DispatchCreate ( IN PDEVICE_OBJECT pDevObj, IN PIRP pIrp );
NTSTATUS DispatchClose ( IN PDEVICE_OBJECT pDevObj, IN PIRP pIrp ) ;
NTSTATUS DispatchWrite ( IN PDEVICE_OBJECT pDevObj, IN PIRP pIrp ) ;
NTSTATUS DispatchRead ( IN PDEVICE_OBJECT pDevObj, IN PIRP pIrp ) ;

extern "C" NTSTATUS DriverEntry ( IN PDRIVER_OBJECT pDrvObj,
								  IN PUNICODE_STRING pRegPath ) 
{

	pDrvObj->DriverUnload = DriverUnload;
	

	pDrvObj->MajorFunction[IRP_MJ_CREATE] = DispatchCreate;
	pDrvObj->MajorFunction[IRP_MJ_CLOSE] = DispatchClose;
	pDrvObj->MajorFunction[IRP_MJ_READ] = DispatchRead;
	pDrvObj->MajorFunction[IRP_MJ_WRITE] = DispatchWrite;
	

	UNICODE_STRING devname ;
	RtlInitUnicodeString(&devname,L"\\Device\\mydevice"); 


	PDEVICE_OBJECT pDevObj ;
	DEVICE_EXTENSION* pDevExt ;
	NTSTATUS status = IoCreateDevice( pDrvObj,
										sizeof(DEVICE_EXTENSION),
										&devname,
										FILE_DEVICE_UNKNOWN,
										0, TRUE, &pDevObj );
	if (!NT_SUCCESS(status))
		return status;


	pDevObj->Flags |= DO_BUFFERED_IO;


	pDevExt = (DEVICE_EXTENSION*)pDevObj->DeviceExtension;

	pDevExt->pBuffer = NULL;
	pDevExt->BufferSize = 0;


	UNICODE_STRING linkname ;
	RtlInitUnicodeString(&linkname,L"\\??\\mydev");

	pDevExt->SymLinkName = linkname;


	status = IoCreateSymbolicLink( &linkname,&devname );
	if (!NT_SUCCESS(status)) 
	{

		IoDeleteDevice( pDevObj );
		return status;
	}

	return STATUS_SUCCESS;
}

void DriverUnload (IN PDRIVER_OBJECT pDrvObj)
{

	PDEVICE_OBJECT pDevObj = pDrvObj->DeviceObject;
	DEVICE_EXTENSION* pDevExt = (DEVICE_EXTENSION*)pDevObj->DeviceExtension;

	if (pDevExt->pBuffer != NULL) 
	{
		ExFreePool(pDevExt->pBuffer);
		pDevExt->pBuffer = NULL;
		pDevExt->BufferSize = 0;
	}
	IoDeleteSymbolicLink(&pDevExt->SymLinkName);
	IoDeleteDevice( pDrvObj->DeviceObject );
	DbgPrint("drv_app1: in driverunload\n");
}

NTSTATUS DispatchCreate (IN PDEVICE_OBJECT	pDevObj,IN PIRP pIrp)
{

	pIrp->IoStatus.Status = STATUS_SUCCESS;
	pIrp->IoStatus.Information = 0;	// no bytes xfered
	IoCompleteRequest( pIrp, IO_NO_INCREMENT );
	return STATUS_SUCCESS;
}

NTSTATUS DispatchClose (IN PDEVICE_OBJECT	pDevObj,IN PIRP	pIrp)
{
	DEVICE_EXTENSION* pDevExt = (DEVICE_EXTENSION*) pDevObj->DeviceExtension;
	if (pDevExt->pBuffer != NULL) 
	{
		ExFreePool(pDevExt->pBuffer);
		pDevExt->pBuffer = NULL;
		pDevExt->BufferSize = 0;
	}
	pIrp->IoStatus.Status = STATUS_SUCCESS;
	pIrp->IoStatus.Information = 0;	// no bytes xfered
	IoCompleteRequest( pIrp, IO_NO_INCREMENT );
	return STATUS_SUCCESS;
}

NTSTATUS DispatchWrite (IN PDEVICE_OBJECT	pDevObj,IN PIRP pIrp)
{
	NTSTATUS status = STATUS_SUCCESS;
	PVOID userBuffer;
	ULONG xferSize;

	PIO_STACK_LOCATION pIrpStack = IoGetCurrentIrpStackLocation( pIrp );

	DEVICE_EXTENSION* pDevExt = (DEVICE_EXTENSION*)	pDevObj->DeviceExtension;

	if (pDevExt->pBuffer != NULL) 
	{
		ExFreePool(pDevExt->pBuffer);
		pDevExt->pBuffer = NULL;
		pDevExt->BufferSize = 0;
	}

	xferSize = pIrpStack->Parameters.Write.Length;
	

	userBuffer = pIrp->AssociatedIrp.SystemBuffer;


	pDevExt->pBuffer =	ExAllocatePool( PagedPool, xferSize );
	if (pDevExt->pBuffer == NULL) 
	{

		status = STATUS_INSUFFICIENT_RESOURCES;
		xferSize = 0;
	}
	else 
	{

		pDevExt->BufferSize = xferSize;
		RtlCopyMemory( pDevExt->pBuffer, userBuffer,xferSize );

		char* p = (char*)pDevExt->pBuffer;
		pDevExt->pBuffer = _strupr(p);
    }

	pIrp->IoStatus.Status = status;
	pIrp->IoStatus.Information = xferSize;	
	IoCompleteRequest( pIrp, IO_NO_INCREMENT );
	return status;
}

NTSTATUS DispatchRead (	IN PDEVICE_OBJECT	pDevObj,IN PIRP	pIrp )
{
	NTSTATUS status = STATUS_SUCCESS;

	PVOID userBuffer;
	ULONG xferSize;

	PIO_STACK_LOCATION pIrpStack = IoGetCurrentIrpStackLocation( pIrp );
	

	DEVICE_EXTENSION* pDevExt = (DEVICE_EXTENSION*)	pDevObj->DeviceExtension;

	xferSize = pIrpStack->Parameters.Read.Length;

	userBuffer = pIrp->AssociatedIrp.SystemBuffer;


	xferSize = (xferSize < pDevExt->BufferSize) ? xferSize : pDevExt->BufferSize;
	

	RtlCopyMemory( userBuffer, pDevExt->pBuffer,xferSize );
	

	ExFreePool( pDevExt->pBuffer );
	pDevExt->pBuffer = NULL;
	pDevExt->BufferSize = 0;


	pIrp->IoStatus.Status = status;
	pIrp->IoStatus.Information = xferSize;	
	IoCompleteRequest( pIrp, IO_NO_INCREMENT );
	return status;
}
