extern "C" 
{
	# include <WDM.h>
}
#include "cstring.h"

enum DEVICE_STATE { Stopped, Started, Removed } ;

struct DEV_EXT 
{
	PDEVICE_OBJECT pDevObj ;
	PDEVICE_OBJECT pLowerDevObj ;
	ULONG DevNum ;
	CString DeviceName ;	
	CString SymLinkName ;	
	PUCHAR Buff ;			
	ULONG BuffSize ;
	ULONG xferCount ;		
	ULONG maxXferCount ;	
	PUCHAR PortBase ;		
	PKINTERRUPT pIntObj ;	
	DEVICE_STATE state ;	
};


# define STS_NOT_IRQ	0x04


# define CTL_NOT_RST	0x04	

# define CTL_INTENB		0x10
# define CTL_DEFAULT	0xC0	


# define ReadStatus(pDevExt) ( READ_PORT_UCHAR ( pDevExt -> PortBase + 1 ) )
# define ReadControl(pDevExt) ( READ_PORT_UCHAR ( pDevExt -> PortBase + 2 ) )
# define WriteControl(pDevExt,bData) ( WRITE_PORT_UCHAR ( pDevExt -> PortBase + 2 ,bData ) )
# define WriteData(pDevExt,bData) ( WRITE_PORT_UCHAR ( pDevExt -> PortBase + 0, bData ) )


void Unload ( PDRIVER_OBJECT pDrvObj ) ;
NTSTATUS AddDevice ( PDRIVER_OBJECT pDrvObj, PDEVICE_OBJECT pdo ) ;
NTSTATUS DispPnp ( PDEVICE_OBJECT pdo, PIRP pIrp ) ;
NTSTATUS PassDownPnP ( PDEVICE_OBJECT pdo, PIRP pIrp ) ;
NTSTATUS StartDev ( PDEVICE_OBJECT pdo, PIRP pIrp ) ;
NTSTATUS StopDev ( PDEVICE_OBJECT pdo, PIRP pIrp );
NTSTATUS RemoveDev (  PDEVICE_OBJECT pdo, PIRP pIrp ) ;
NTSTATUS Create ( PDEVICE_OBJECT pDevObj, PIRP pIrp ) ;
NTSTATUS Close ( PDEVICE_OBJECT pDevObj, PIRP pIrp ) ;
NTSTATUS Write ( PDEVICE_OBJECT pDevObj,  PIRP pIrp ) ;
NTSTATUS Read ( PDEVICE_OBJECT pDevObj, PIRP pIrp ) ;
BOOLEAN Isr ( PKINTERRUPT pIntObj, PVOID pServiceContext ) ;
BOOLEAN TransmitByte ( PVOID pArg ) ;
void StartIo ( PDEVICE_OBJECT pDevObj,  PIRP pIrp ) ;
void DpcForIsr ( PKDPC pDpc, PDEVICE_OBJECT pDevObj, PIRP pIrp, PVOID pContext ) ;

extern "C" NTSTATUS DriverEntry (  PDRIVER_OBJECT pDrvObj, PUNICODE_STRING pRegPath ) 
{
	DbgPrint ( "LoopBack: DriverEntry\n" ) ;
	NTSTATUS status = STATUS_SUCCESS ;
	pDrvObj -> DriverUnload = Unload ;
	pDrvObj -> DriverExtension -> AddDevice = AddDevice ; 
	pDrvObj -> MajorFunction [ IRP_MJ_PNP ] = DispPnp ; 
	pDrvObj -> MajorFunction [ IRP_MJ_CREATE ] = Create ;
	pDrvObj -> MajorFunction [ IRP_MJ_CLOSE ] =  Close ;
	pDrvObj -> MajorFunction [ IRP_MJ_WRITE ] = Write ;
	pDrvObj -> MajorFunction [ IRP_MJ_READ ] = Read ;
	pDrvObj -> DriverStartIo = StartIo ;
	
	return status ;
}

NTSTATUS AddDevice (  PDRIVER_OBJECT pDrvObj, PDEVICE_OBJECT pdo ) 
{
	NTSTATUS status ;
	PDEVICE_OBJECT pfdo ;
	DEV_EXT* pDevExt ;
	static int DevNum ;

	DbgPrint ( "LoopBack: AddDevice() current DevNum = %d\n", DevNum ) ;
	

	CString devName ("\\Device\\LoopBack" ) ; 
	devName += CString ( DevNum ) ;


	status = IoCreateDevice ( pDrvObj,
							  sizeof ( DEV_EXT ),
							  &( UNICODE_STRING ) devName,
							  FILE_DEVICE_UNKNOWN,
							  0,
							  FALSE,
							  &pfdo ) ;
	if ( ! NT_SUCCESS ( status ) )
		return status ;

	pfdo -> Flags |= DO_BUFFERED_IO ; 

	
	pDevExt = (DEV_EXT*) pfdo -> DeviceExtension ;
	pDevExt -> pDevObj = pfdo ;	// back pointer
	pDevExt -> DevNum = DevNum ;
	pDevExt -> DeviceName = devName ;
	pDevExt -> pIntObj = NULL ;
	pDevExt -> state = Stopped ;

	
	pDevExt -> pLowerDevObj = IoAttachDeviceToDeviceStack ( pfdo, pdo ) ; 

	
	CString symLinkName ( "\\??\\LOOP" ) ;
	symLinkName += CString ( DevNum + 1 ) ;	
	pDevExt -> SymLinkName = symLinkName ;

	
	status = IoCreateSymbolicLink ( &(UNICODE_STRING)symLinkName, &(UNICODE_STRING) devName ) ;
	if (! NT_SUCCESS ( status ) )  
	{
		IoDeleteDevice ( pfdo ) ; 
		return status ;
	}

	
	IoInitializeDpcRequest ( pfdo, DpcForIsr ) ;

    
    pfdo -> Flags &= ~DO_DEVICE_INITIALIZING ;

	
	DevNum ++ ;
	return STATUS_SUCCESS ;
}

NTSTATUS DispPnp ( PDEVICE_OBJECT pdo, PIRP pIrp ) 
{
	PIO_STACK_LOCATION pIrpStack ;
	pIrpStack = IoGetCurrentIrpStackLocation ( pIrp ) ; 

	DbgPrint ( "LoopBack: Received PNP IRP: %d\n", pIrpStack -> MinorFunction ) ;

	switch ( pIrpStack -> MinorFunction ) 
	{
		case IRP_MN_START_DEVICE :
			return StartDev ( pdo, pIrp ) ;
		case IRP_MN_STOP_DEVICE :
			return StopDev ( pdo, pIrp ) ;
		case IRP_MN_REMOVE_DEVICE :
			return RemoveDev ( pdo, pIrp ) ;
		default:
			return PassDownPnP ( pdo, pIrp ) ;
	}
}

NTSTATUS PassDownPnP ( PDEVICE_OBJECT pdo, PIRP pIrp ) 
{
	IoSkipCurrentIrpStackLocation ( pIrp ) ;
	DEV_EXT* pDevExt = (DEV_EXT*) pdo -> DeviceExtension ;
	return IoCallDriver ( pDevExt -> pLowerDevObj, pIrp ) ;
}

NTSTATUS StartDev ( PDEVICE_OBJECT pdo, PIRP pIrp ) 
{
	PIO_STACK_LOCATION pIrpStack = IoGetCurrentIrpStackLocation ( pIrp ) ;
	DEV_EXT* pDevExt = (DEV_EXT*) pdo -> DeviceExtension ; 
	DbgPrint ("LoopBack: StartDevice, Symbolic device #%d\n", pDevExt -> DevNum + 1 ) ;

	PCM_RESOURCE_LIST pResList ;
	PCM_FULL_RESOURCE_DESCRIPTOR pFullDesc ;
	PCM_PARTIAL_RESOURCE_LIST pPartialList ;
	PCM_PARTIAL_RESOURCE_DESCRIPTOR pPartialDesc ;
	
	int i ;
	NTSTATUS status ;

	KIRQL IRQL ;
	ULONG Vector ;
	KAFFINITY Affinity ;
	ULONG  PortLength ;

	pResList =	pIrpStack -> Parameters.StartDevice.AllocatedResourcesTranslated ;
	pFullDesc = pResList -> List ;
	pPartialList = &pFullDesc -> PartialResourceList ;

	for ( i = 0 ; i < (int) pPartialList -> Count; i++ ) 
	{
		pPartialDesc = &pPartialList -> PartialDescriptors [ i ] ;
		switch ( pPartialDesc -> Type )  
		{
			case CmResourceTypeInterrupt :
				IRQL = (KIRQL) pPartialDesc -> u.Interrupt.Level ;
				Vector =  pPartialDesc -> u.Interrupt.Vector ;
				Affinity =  pPartialDesc -> u.Interrupt.Affinity ;
				DbgPrint ("LoopBack: Claiming Interrupt Resources: IRQL=%d Vector=0x%03X Affinity=%X\n",IRQL, Vector, Affinity ) ;
			break ;
		case CmResourceTypePort:
			if( pPartialDesc -> u.Port.Start.LowPart !=0 )
			{
				pDevExt -> PortBase = (PUCHAR)pPartialDesc -> u.Port.Start.LowPart ;
				PortLength = pPartialDesc -> u.Port.Length ;
				DbgPrint ( "LoopBack: Claiming Port Resources: PortBase=%X Len=%d\n", pDevExt->PortBase, PortLength ) ;
			}
			break;
		
		}
	}
	if ( IRQL == 0 || pDevExt -> PortBase == 0 )
	{
		DbgPrint ( "returning back interrupt not connected\n") ;
		return STATUS_BIOS_FAILED_TO_CONNECT_INTERRUPT ;
	}
	
	status = IoConnectInterrupt ( &pDevExt -> pIntObj,	
								  Isr,				
								  pDevExt,			
								  NULL,				
								  Vector,	
								  IRQL,	
								  IRQL,	
								  Latched,	
								  TRUE,		
								  Affinity,	
								  FALSE ) ;	
	if ( ! NT_SUCCESS ( status ) ) 
	{
		DbgPrint("LoopBack: Interrupt connection failure: %X\n", status ) ;
		return status ;
	}
	DbgPrint ("LoopBack: Interrupt successfully connected\n") ;
	pDevExt -> state = Started ;
	return PassDownPnP ( pdo, pIrp ) ;
}

NTSTATUS StopDev ( PDEVICE_OBJECT pdo, PIRP pIrp ) 
{
	DbgPrint ("LoopBack: StopDevice Handler\n" ) ;
	DEV_EXT* pDevExt = (DEV_EXT*) pdo -> DeviceExtension ;
	
	if (pDevExt->pIntObj)
		IoDisconnectInterrupt ( pDevExt -> pIntObj ) ; 

	pDevExt -> state = Stopped ;
	return PassDownPnP ( pdo, pIrp ) ;
}

NTSTATUS RemoveDev ( PDEVICE_OBJECT pdo, PIRP pIrp ) 
{
	DbgPrint ( "LoopBack: RemoveDevice Handler\n" ) ;
	DEV_EXT* pDevExt = (DEV_EXT*) pdo -> DeviceExtension ;

	if ( pDevExt -> state == Started ) 
	{
		if ( pDevExt -> pIntObj )
			IoDisconnectInterrupt ( pDevExt -> pIntObj ) ;
	}
	
	
	UNICODE_STRING pLinkName = pDevExt -> SymLinkName ;
	
	IoDeleteSymbolicLink ( &pLinkName ) ;

	DbgPrint ( "PPORT: Symbolic Link MPMP%d Deleted\n", pDevExt -> DevNum + 1 ) ;
	
	
	IoDeleteDevice ( pdo ) ;

	pDevExt -> state = Removed ;
	return PassDownPnP ( pdo, pIrp ) ;
}


void Unload ( PDRIVER_OBJECT pDrvObj ) 
{
	DbgPrint ( "LoopBack: Unload\n" ) ;
}

NTSTATUS Create ( PDEVICE_OBJECT pDevObj, PIRP pIrp ) 
{
	DbgPrint ( "LoopBack: Create\n" ) ;
	DEV_EXT* pDevExt = (DEV_EXT*) pDevObj -> DeviceExtension ;
	
	NTSTATUS status = STATUS_SUCCESS ;
	
	if (pDevExt -> state != Started )
		status = STATUS_DEVICE_REMOVED ;
	
	pIrp -> IoStatus.Status = status ;
	pIrp -> IoStatus.Information = 0 ;	
	IoCompleteRequest ( pIrp, IO_NO_INCREMENT ) ;
	return status ;
}

NTSTATUS Close ( PDEVICE_OBJECT	pDevObj, PIRP pIrp ) 
{
	DbgPrint ( "LoopBack: Close\n" ) ;
	
	DEV_EXT* pDevExt = (DEV_EXT*) pDevObj -> DeviceExtension ; 
	if ( pDevExt -> Buff != NULL ) 
	{
		ExFreePool ( pDevExt -> Buff ) ;
		pDevExt -> Buff = NULL ;
		pDevExt -> BuffSize = 0 ; 
	}
	pIrp -> IoStatus.Status = STATUS_SUCCESS ;
	pIrp -> IoStatus.Information = 0 ;	
	IoCompleteRequest ( pIrp, IO_NO_INCREMENT ) ;
	return STATUS_SUCCESS ;
}

void Cancel ( PDEVICE_OBJECT pDevObj, PIRP pIrp )
{
	DbgPrint ("LoopBack: IRP Canceled\n" ) ;


	pIrp -> IoStatus.Status = STATUS_CANCELLED ;
	pIrp -> IoStatus.Information = 0 ;	
	IoCompleteRequest ( pIrp, IO_NO_INCREMENT ) ;
	IoStartNextPacket ( pDevObj, TRUE ) ;
}

NTSTATUS Write ( PDEVICE_OBJECT	pDevObj, PIRP pIrp ) 
{
	DbgPrint ("LoopBack: Write Operation requested (Write)\n") ;
	IoMarkIrpPending ( pIrp ) ;
	IoStartPacket ( pDevObj, pIrp, 0, Cancel ) ;
	return STATUS_PENDING ;
}

NTSTATUS Read ( PDEVICE_OBJECT pDevObj, PIRP pIrp ) 
{
	DbgPrint ( "LoopBack: Read Operation requested (Read)\n" ) ;
	NTSTATUS status = STATUS_SUCCESS ;
	PVOID userBuffer ;
	ULONG xferSize ;
	
	PIO_STACK_LOCATION pIrpStack = IoGetCurrentIrpStackLocation ( pIrp ) ; 
	DEV_EXT* pDevExt = (DEV_EXT*) pDevObj -> DeviceExtension ; 
	xferSize = pIrpStack -> Parameters.Read.Length ; 
	userBuffer = pIrp -> AssociatedIrp.SystemBuffer ; 

	xferSize = ( xferSize < pDevExt -> BuffSize ) ? xferSize : pDevExt -> BuffSize ; 
	RtlCopyMemory ( userBuffer, pDevExt -> Buff, xferSize ) ; 
	ExFreePool ( pDevExt -> Buff ) ; 
	pDevExt -> Buff = NULL ;
	pDevExt -> BuffSize = 0 ;


	pIrp -> IoStatus.Status = status ;
	pIrp -> IoStatus.Information = xferSize ;	
	IoCompleteRequest ( pIrp, IO_NO_INCREMENT ) ;
	return status ;
}

BOOLEAN Isr ( PKINTERRUPT pIntObj, PVOID pServiceContext ) 
{
	DbgPrint ( "LoopBack: inside isr\n" ) ;

	DEV_EXT* pDevExt = (DEV_EXT*) pServiceContext ;
	PDEVICE_OBJECT pDevObj = pDevExt -> pDevObj ;
	PIRP pIrp = pDevObj -> CurrentIrp ;

	UCHAR status = ReadStatus ( pDevExt ) ;
	if ( !( status & STS_NOT_IRQ ) )
	{
		DbgPrint("LoopBack: returning false");
		return FALSE;
	}

	WriteControl ( pDevExt, CTL_DEFAULT ) ;
	

	if ( ! TransmitByte ( pDevExt  ) )
		IoRequestDpc ( pDevObj, pIrp, (PVOID) pDevExt ) ; 

	return TRUE ;
}

BOOLEAN TransmitByte ( PVOID pArg ) 
{
	DEV_EXT* pDevExt = ( DEV_EXT* ) pArg ;
	 
	if ( pDevExt -> xferCount >= pDevExt -> maxXferCount )
		return FALSE ;

	PDEVICE_OBJECT pDevObj = pDevExt -> pDevObj ;
	PIRP pIrp = pDevObj -> CurrentIrp ; 
	

	PUCHAR userBuffer = (PUCHAR) pIrp -> AssociatedIrp.SystemBuffer ;
	UCHAR nextByte = userBuffer [ pDevExt -> xferCount ] ;


	DbgPrint ("LoopBack: TransmitByte: Sending 0x%02X to port %X\n", nextByte, pDevExt -> PortBase ) ;

	WriteData ( pDevExt, nextByte & 1 ) ;
	UCHAR bits = ( nextByte & 0x8 ) + ( ( nextByte & 0x6 ) >> 1 ) ;
	bits ^= 0x3 ; 
	
	
	WriteControl ( pDevExt, bits | CTL_DEFAULT ) ;

	
	UCHAR status = ReadStatus ( pDevExt ) ;
	DbgPrint ("LoopBack: status check 0x%03X\n", status ) ;

	
	UCHAR readByte =  ( ( status & 0x8 ) >> 3 ) | (( status & 0x30 ) >> 3 ) | ( ( status & 0x80 ) >> 4 ) ;

	// rotate bits to right by 1
	readByte = readByte >> 1 ;

	pDevExt -> Buff [ pDevExt -> xferCount ++ ] = readByte  ;
	DbgPrint ("LoopBack: TransmitByte read character: 0x%03X\n", readByte ) ;


	DbgPrint ("LoopBack: TransmitByte: generating interrupt.\n" ) ;

	WriteControl ( pDevExt, CTL_INTENB | CTL_DEFAULT ) ;

	WriteControl ( pDevExt, CTL_INTENB | CTL_NOT_RST | CTL_DEFAULT ) ;
	

	KeStallExecutionProcessor ( 50 ) ;
	
	WriteControl ( pDevExt, CTL_INTENB | CTL_DEFAULT ) ;
	return TRUE ;
}


void StartIo ( PDEVICE_OBJECT pDevObj, PIRP pIrp ) 
{
	DbgPrint ("LoopBack: Start I/O Operation\n" ) ;
	PIO_STACK_LOCATION	pIrpStack = IoGetCurrentIrpStackLocation ( pIrp ) ;
	DEV_EXT* pDevExt = (DEV_EXT*) pDevObj->DeviceExtension ;
	PUCHAR userBuffer ;
	ULONG xferSize ;
	
	switch( pIrpStack -> MajorFunction ) 
	{
		
		case IRP_MJ_WRITE :
			
			pDevExt -> maxXferCount =  pIrpStack -> Parameters.Write.Length ;
			pDevExt -> xferCount = 0 ;
			
			if ( pDevExt -> Buff != NULL ) 
			{
				ExFreePool ( pDevExt -> Buff ) ;
				pDevExt -> Buff = NULL ;
				pDevExt -> BuffSize = 0 ;
			}
			
			xferSize = pIrpStack -> Parameters.Write.Length ;
			
			
			userBuffer = (PUCHAR) pIrp -> AssociatedIrp.SystemBuffer ;
			
			pDevExt -> Buff = (PUCHAR) ExAllocatePool ( PagedPool, xferSize ) ;
			if ( pDevExt -> Buff == NULL ) 
			{
			
				pIrp -> IoStatus.Status = STATUS_INSUFFICIENT_RESOURCES ;
				pIrp -> IoStatus.Information = 0 ;
				IoCompleteRequest ( pIrp, IO_NO_INCREMENT ) ;
				IoStartNextPacket ( pDevObj, FALSE ) ;
			}
			pDevExt -> BuffSize = xferSize ;
			
			
			DbgPrint ( "LoopBack: StartIO: Transmitting first byte of %d\n", pDevExt -> BuffSize ) ;
			
			if ( ! KeSynchronizeExecution ( pDevExt -> pIntObj, TransmitByte, pDevExt ) )
				DpcForIsr ( NULL,	pDevObj, pIrp, pDevExt ) ;
			break;
			
			default: 
				pIrp -> IoStatus.Status = STATUS_NOT_SUPPORTED ;
				pIrp -> IoStatus.Information = 0 ;
				IoCompleteRequest ( pIrp, IO_NO_INCREMENT ) ;
				IoStartNextPacket ( pDevObj, FALSE ) ;
				break;
	}
}

void DpcForIsr ( PKDPC pDpc, PDEVICE_OBJECT pDevObj, PIRP pIrp, PVOID pContext )
{
	DEV_EXT* pDevExt = (DEV_EXT*) pContext ;
	DbgPrint ( "LoopBack: DpcForIsr, xferCount = %d\n", pDevExt -> xferCount ) ;
	pIrp -> IoStatus.Information = pDevExt -> xferCount ;
	pIrp -> IoStatus.Status = STATUS_SUCCESS ; 

	
	if( pDpc == NULL )
		IoCompleteRequest ( pIrp, IO_NO_INCREMENT ) ;
	else
		IoCompleteRequest ( pIrp, IO_PARALLEL_INCREMENT ) ;

	IoStartNextPacket ( pDevObj, FALSE ) ; 
}