extern "C"
{
	#include <wdm.h>
	#pragma warning(disable:4200) 
	#include <usbdi.h>
	#include <usbdlib.h>
	#include <stdio.h>
	#include <usb100.h>
	#include "str.h"
}	

struct DEV_EXT 
{
	PDEVICE_OBJECT pDevObj ;				
	PDEVICE_OBJECT pLowerDevObj ;			
	ULONG DevNum ;
	CString DeviceName ;					
	CString SymLinkName ;					
	USB_DEVICE_DESCRIPTOR dd;				
	PUSB_CONFIGURATION_DESCRIPTOR pcd;		
	LANGID langid;							
};

#define MSGUSBSTRING(d,s,i) { \
		UNICODE_STRING sd; \
		if (i && NT_SUCCESS(GetStringDescriptor(d,i,&sd))) { \
			DbgPrint(s, sd.Buffer); \
			RtlFreeUnicodeString(&sd); \
		}}

void Unload ( PDRIVER_OBJECT pDrvObj ) ;
NTSTATUS GetStringDescriptor(PDEVICE_OBJECT fdo, UCHAR istring, PUNICODE_STRING s);
NTSTATUS SendUrb(PDEVICE_OBJECT pfdo, PURB urb);
NTSTATUS Close ( PDEVICE_OBJECT	pDevObj, PIRP pIrp ) ;
NTSTATUS Create ( PDEVICE_OBJECT pDevObj, PIRP pIrp );
NTSTATUS RemoveDev ( PDEVICE_OBJECT pfdo, PIRP pIrp );
NTSTATUS StopDev ( PDEVICE_OBJECT pfdo, PIRP pIrp ) ;
NTSTATUS StartDev ( PDEVICE_OBJECT pfdo, PIRP pIrp ); 
NTSTATUS DispPnp ( PDEVICE_OBJECT pfdo, PIRP pIrp );
NTSTATUS CompleteRequest(PIRP pIrp, NTSTATUS status, ULONG_PTR info);
NTSTATUS AddDevice (  PDRIVER_OBJECT pDrvObj, PDEVICE_OBJECT ppdo ) ;

#pragma code_seg("init")
extern "C" NTSTATUS DriverEntry(PDRIVER_OBJECT pDrvObj, PUNICODE_STRING pRegPath) 
{
	DbgPrint ("minusb: DriverEntry( )\n");
	pDrvObj -> DriverUnload = Unload ;
	pDrvObj -> DriverExtension -> AddDevice = AddDevice ;
	pDrvObj -> MajorFunction [ IRP_MJ_CREATE ] = Create ;
	pDrvObj -> MajorFunction [ IRP_MJ_CLOSE ] =  Close ;
	pDrvObj -> MajorFunction [ IRP_MJ_PNP ] = DispPnp ; 
	return STATUS_SUCCESS ;
}

#pragma code_seg("page")
void Unload ( PDRIVER_OBJECT pDrvObj ) 
{
	DbgPrint("minusb: Unload( )\n");
}

#pragma code_seg("page")
NTSTATUS AddDevice (  PDRIVER_OBJECT pDrvObj, PDEVICE_OBJECT ppdo ) 
{
	static int DevNum ;
	CString devName ("\\Device\\MinimalUSB" ) ; 
	devName += CString ( DevNum ) ;

	PDEVICE_OBJECT pfdo ;
	NTSTATUS status = IoCreateDevice ( pDrvObj,
							  sizeof ( DEV_EXT ),
							  &(UNICODE_STRING)devName,
							  FILE_DEVICE_UNKNOWN,
							  0,
							  FALSE,
							  &pfdo ) ;
	if ( ! NT_SUCCESS ( status ) )
		return status ;

	pfdo -> Flags |= DO_POWER_PAGABLE ; 

	DEV_EXT* pdx = (DEV_EXT*) pfdo -> DeviceExtension ;
	
	pdx -> pDevObj = pfdo ;	
	pdx -> DeviceName = devName ;
	pdx -> pLowerDevObj = IoAttachDeviceToDeviceStack ( pfdo, ppdo ) ; 

	CString symLinkName ( "\\??\\MINUSB" ) ;
	symLinkName += CString ( DevNum + 1 ) ;	
	pdx -> SymLinkName = symLinkName ;

	status = IoCreateSymbolicLink ( &(UNICODE_STRING)symLinkName, &(UNICODE_STRING)devName ) ;
	if (! NT_SUCCESS ( status ) )  
	{
		IoDeleteDevice ( pfdo ) ;
		return status ;
	}
    pfdo -> Flags &= ~DO_DEVICE_INITIALIZING ;
	DevNum ++ ;
	return STATUS_SUCCESS ;
}

#pragma code_seg()
NTSTATUS CompleteRequest(PIRP pIrp, NTSTATUS status, ULONG_PTR info)
{							
	pIrp->IoStatus.Status = status;
	pIrp->IoStatus.Information = info;
	IoCompleteRequest(pIrp, IO_NO_INCREMENT);
	return status;
}							

#pragma code_seg("page")
NTSTATUS DispPnp ( PDEVICE_OBJECT pfdo, PIRP pIrp )
{
	DEV_EXT* pdx = (DEV_EXT*)pfdo->DeviceExtension ;
	
	PIO_STACK_LOCATION pIrpStack = IoGetCurrentIrpStackLocation ( pIrp ) ; 
	NTSTATUS status ;

	switch ( pIrpStack -> MinorFunction ) 
	{
		case IRP_MN_START_DEVICE :
			status =  StartDev ( pfdo, pIrp ) ;
			break;
		case IRP_MN_STOP_DEVICE :
			status = StopDev ( pfdo, pIrp ) ;
			break ;
		case IRP_MN_REMOVE_DEVICE :
			status = RemoveDev ( pfdo, pIrp ) ;
			break ;
		default:
			IoSkipCurrentIrpStackLocation(pIrp);
			status = IoCallDriver(pdx->pLowerDevObj, pIrp);
	}
	return status;
}

#pragma code_seg("page")
NTSTATUS StartDev ( PDEVICE_OBJECT pfdo, PIRP pIrp ) 
{
	PIO_STACK_LOCATION pIrpStack = IoGetCurrentIrpStackLocation ( pIrp ) ;
	DEV_EXT* pdx = (DEV_EXT*) pfdo -> DeviceExtension ; 
	
	URB urb;		
	UsbBuildGetDescriptorRequest(&urb, sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST), USB_DEVICE_DESCRIPTOR_TYPE,	0, 0, &pdx->dd, NULL, sizeof(pdx->dd), NULL);
	NTSTATUS status = SendUrb(pfdo, &urb);
	if (!NT_SUCCESS(status))
	{
		DbgPrint("minusb: - Error trying to read device descriptor\n");
		return status;
	}
	MSGUSBSTRING(pfdo,  " - Configuring device from %ws\n", pdx->dd.iManufacturer);
	MSGUSBSTRING(pfdo,  " - Product is %ws\n", pdx->dd.iProduct);
	MSGUSBSTRING(pfdo,  " - Serial number is %ws\n", pdx->dd.iSerialNumber);
	DbgPrint("minusb: Total number of configs = %d\n", pdx->dd.bNumConfigurations );


	_USB_CONFIGURATION_DESCRIPTOR tcd;
	UsbBuildGetDescriptorRequest(&urb, sizeof(_URB_CONTROL_DESCRIPTOR_REQUEST), USB_CONFIGURATION_DESCRIPTOR_TYPE,0, 0, &tcd, NULL, sizeof(tcd), NULL);
	status = SendUrb(pfdo, &urb);
	if (!NT_SUCCESS(status))
	{
		DbgPrint( "minusb: Error %X trying to read configuration descriptor 1\n", status);
		return status;
	}
	ULONG size = tcd.wTotalLength;
	PUSB_CONFIGURATION_DESCRIPTOR pcd = (PUSB_CONFIGURATION_DESCRIPTOR) ExAllocatePool(NonPagedPool, size);
	if (!pcd)
	{
		DbgPrint( "minusb: Unable to allocate %X bytes for configuration descriptor\n", size);
		return STATUS_INSUFFICIENT_RESOURCES;
	}
	__try
	{
		UsbBuildGetDescriptorRequest(&urb, sizeof(_URB_CONTROL_DESCRIPTOR_REQUEST), USB_CONFIGURATION_DESCRIPTOR_TYPE,	0, 0, pcd, NULL, size, NULL);
		status = SendUrb(pfdo, &urb);
		if (!NT_SUCCESS(status))
		{
			DbgPrint("minusb: Error %X trying to read configuration descriptor 1\n", status);
			return status;
		}
    	MSGUSBSTRING(pfdo, "- Selecting configuration named %ws\n", pcd->iConfiguration); 
		PUSB_INTERFACE_DESCRIPTOR pid = USBD_ParseConfigurationDescriptorEx(pcd, pcd,-1, -1, -1, -1, -1);
		if (!pid)
		{
			DbgPrint("minusb: No interface for device\n");
			return STATUS_DEVICE_CONFIGURATION_ERROR;
		}
		MSGUSBSTRING(pfdo,  " - Selecting interface named %ws\n", pid->iInterface);
		USBD_INTERFACE_LIST_ENTRY interfaces[2] = 
		{
			{pid, NULL},
			{NULL, NULL},		
		};
		PURB selurb = USBD_CreateConfigurationRequestEx(pcd, interfaces);
		if (!selurb)
		{
			DbgPrint("minusb: Unable to create configuration request\n");
			return STATUS_INSUFFICIENT_RESOURCES;
		}
		__try
		{

			PUSBD_INTERFACE_INFORMATION pii = interfaces[0].Interface;
			status = SendUrb(pfdo, selurb);
			if (!NT_SUCCESS(status))
			{
				DbgPrint("minusb: Error %X trying to select configuration\n", status);
				return status;
			}
			pdx->pcd = pcd;
			pcd = NULL;
		}
		__finally
		{
			ExFreePool(selurb);
		}
	}
	__finally
	{
		if (pcd)
			ExFreePool(pcd);
	}
	pIrp -> IoStatus.Status = STATUS_SUCCESS ;
	pIrp -> IoStatus.Information = 0 ;	
	IoCompleteRequest ( pIrp, IO_NO_INCREMENT ) ;
	return STATUS_SUCCESS ;
}

NTSTATUS StopDev ( PDEVICE_OBJECT pfdo, PIRP pIrp ) 
{
	DbgPrint ("minusb: StopDevice Handler\n" ) ;
	DEV_EXT* pe = (DEV_EXT*) pfdo->DeviceExtension;

	
	URB urb;
	UsbBuildSelectConfigurationRequest(&urb, sizeof(_URB_SELECT_CONFIGURATION), NULL);
	NTSTATUS status = SendUrb(pfdo, &urb);
	if (!NT_SUCCESS(status))
	{
		DbgPrint("minusb: Error %X trying to deconfigure device\n", status);
	}
	
	if (pe->pcd)
		ExFreePool(pe->pcd);
	pe->pcd = NULL;
	
	pIrp -> IoStatus.Status = STATUS_SUCCESS ;
	pIrp -> IoStatus.Information = 0 ;	
	IoCompleteRequest ( pIrp, IO_NO_INCREMENT ) ;
	return STATUS_SUCCESS ;
}


NTSTATUS RemoveDev ( PDEVICE_OBJECT pfdo, PIRP pIrp ) 
{
	DbgPrint ( "minusb: RemoveDevice Handler\n" ) ;

	DEV_EXT* pdx = (DEV_EXT*) pfdo -> DeviceExtension ;

	UNICODE_STRING pLinkName = pdx -> SymLinkName ;
	IoDeleteSymbolicLink ( &pLinkName ) ;

	if (pdx->pLowerDevObj)
		IoDetachDevice(pdx->pLowerDevObj);

	IoDeleteDevice ( pfdo ) ;
	
	pIrp -> IoStatus.Status = STATUS_SUCCESS ;
	pIrp -> IoStatus.Information = 0 ;	
	IoCompleteRequest ( pIrp, IO_NO_INCREMENT ) ;
	return STATUS_SUCCESS ;
}

NTSTATUS Create ( PDEVICE_OBJECT pDevObj, PIRP pIrp ) 
{
	DEV_EXT* pdx ;
	NTSTATUS status =STATUS_SUCCESS;

	DbgPrint ( "minusb: Create\n" ) ;
	pdx = (DEV_EXT*) pDevObj -> DeviceExtension ;
	
	pIrp -> IoStatus.Status = status ;
	pIrp -> IoStatus.Information = 0 ;	// no bytes xfered
	IoCompleteRequest ( pIrp, IO_NO_INCREMENT ) ;
	return status ;
}

NTSTATUS Close ( PDEVICE_OBJECT	pfdo, PIRP pIrp ) 
{
	DbgPrint ( "minusb: Close\n" ) ;
	DEV_EXT* pdx = (DEV_EXT*) pfdo -> DeviceExtension ;
	pIrp -> IoStatus.Status = STATUS_SUCCESS ;
	pIrp -> IoStatus.Information = 0 ;	
	IoCompleteRequest ( pIrp, IO_NO_INCREMENT ) ;
	return STATUS_SUCCESS ;
}

NTSTATUS SendUrb(PDEVICE_OBJECT pfdo, PURB urb)
{
	KEVENT event;
	IO_STATUS_BLOCK iostatus;
	PIRP Irp ;
	PIO_STACK_LOCATION stack ;
	NTSTATUS status ;

	DEV_EXT* pdx = (DEV_EXT*) pfdo->DeviceExtension;
	KeInitializeEvent(&event, NotificationEvent, FALSE);
	Irp = IoBuildDeviceIoControlRequest(IOCTL_INTERNAL_USB_SUBMIT_URB,pdx->pLowerDevObj, NULL, 0, NULL, 0, TRUE, &event, &iostatus);
	if (!Irp)
	{
		DbgPrint("minusb: Unable to allocate IRP for sending URB\n");
		return STATUS_INSUFFICIENT_RESOURCES;
	}
	stack = IoGetNextIrpStackLocation(Irp);
	stack->Parameters.Others.Argument1 = (PVOID) urb;
	status = IoCallDriver(pdx->pLowerDevObj, Irp);
	if (status == STATUS_PENDING)
	{
		KeWaitForSingleObject(&event, Executive, KernelMode, FALSE, NULL);
		status = iostatus.Status;
	}
	return status;
}	

NTSTATUS GetStringDescriptor(PDEVICE_OBJECT fdo, UCHAR istring, PUNICODE_STRING s)
{							
	NTSTATUS status;
	DEV_EXT* pe = (DEV_EXT*) fdo->DeviceExtension;
	URB urb;
	UCHAR data[256];

	if (!pe->langid)
	{						
		UsbBuildGetDescriptorRequest(&urb, sizeof(_URB_CONTROL_DESCRIPTOR_REQUEST), USB_STRING_DESCRIPTOR_TYPE,0, 0, data, NULL, sizeof(data), NULL);
		status = SendUrb(fdo, &urb);
		if (!NT_SUCCESS(status))
			return status;
		pe->langid = *(LANGID*)(data + 2); 
	}						
	UsbBuildGetDescriptorRequest(&urb, sizeof(_URB_CONTROL_DESCRIPTOR_REQUEST), USB_STRING_DESCRIPTOR_TYPE,istring, pe->langid, data, NULL, sizeof(data), NULL);
	status = SendUrb(fdo, &urb);
	if (!NT_SUCCESS(status))
		return status;

	ULONG nchars = (data[0] - 2) / 2;
	PWSTR p = (PWSTR) ExAllocatePool(PagedPool, data[0]);
	if (!p)
		return STATUS_INSUFFICIENT_RESOURCES;

	memcpy(p, data + 2, nchars*2);
	p[nchars] = 0;

	s->Length = (USHORT) (2 * nchars);
	s->MaximumLength = (USHORT) ((2 * nchars) + 2);
	s->Buffer = p;

	return STATUS_SUCCESS;
}	